/*
 * Copyright (c) 2016-2020, National Institute of Informatics
 */
/*
 * Licensed to the University Corporation for Advanced Internet Development,
 * Inc. (UCAID) under one or more contributor license agreements.  See the
 * NOTICE file distributed with this work for additional information regarding
 * copyright ownership. The UCAID licenses this file to You under the Apache
 * License, Version 2.0 (the "License"); you may not use this file except in
 * compliance with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package jp.gakunin.idp.attribute.filter.impl;

import com.google.common.base.Strings;
import com.google.common.collect.HashMultimap;
import com.google.common.collect.Multimap;

import java.util.Collection;
import java.util.List;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.function.Function;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.annotation.concurrent.ThreadSafe;

import jp.gakunin.idp.attribute.filter.context.AttributeFilterOptionalContext;
import jp.gakunin.idp.attribute.filter.matcher.saml.impl.AttributeInMetadataMatcher;
import jp.gakunin.idp.attribute.UApproveJPIdPAttribute;

import org.opensaml.messaging.context.navigate.ChildContextLookup;
import org.opensaml.messaging.context.navigate.RootContextLookup;
import org.opensaml.profile.context.MetricContext;
import org.opensaml.profile.context.ProfileRequestContext;
import org.opensaml.saml.common.messaging.context.AttributeConsumingServiceContext;
import org.opensaml.saml.common.messaging.context.SAMLMetadataContext;
import org.opensaml.saml.saml2.metadata.AttributeConsumingService;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import net.shibboleth.ext.spring.service.AbstractServiceableComponent;
import net.shibboleth.idp.attribute.AttributeEncodingException;
import net.shibboleth.idp.attribute.IdPAttribute;
import net.shibboleth.idp.attribute.IdPAttributeValue;
import net.shibboleth.idp.attribute.filter.AttributeFilter;
import net.shibboleth.idp.attribute.filter.AttributeFilterException;
import net.shibboleth.idp.attribute.filter.AttributeFilterPolicy;
import net.shibboleth.idp.attribute.filter.AttributeRule;
import net.shibboleth.idp.attribute.filter.Matcher;
import net.shibboleth.idp.attribute.filter.matcher.logic.impl.AbstractComposedMatcher;
import net.shibboleth.idp.attribute.filter.matcher.logic.impl.AndMatcher;
import net.shibboleth.idp.attribute.filter.matcher.logic.impl.OrMatcher;
import net.shibboleth.idp.attribute.filter.context.AttributeFilterContext;
import net.shibboleth.idp.attribute.filter.context.AttributeFilterWorkContext;
import net.shibboleth.idp.attribute.transcoding.AttributeTranscoderRegistry;
import net.shibboleth.idp.attribute.transcoding.TranscodingRule;
import net.shibboleth.idp.saml.attribute.transcoding.SAML1AttributeTranscoder;
import net.shibboleth.idp.saml.attribute.transcoding.SAML2AttributeTranscoder;
import net.shibboleth.idp.saml.xml.SAMLConstants;
import net.shibboleth.utilities.java.support.annotation.constraint.NonnullElements;
import net.shibboleth.utilities.java.support.annotation.constraint.NotEmpty;
import net.shibboleth.utilities.java.support.annotation.constraint.NullableElements;
import net.shibboleth.utilities.java.support.annotation.constraint.Unmodifiable;
import net.shibboleth.utilities.java.support.component.ComponentInitializationException;
import net.shibboleth.utilities.java.support.component.ComponentSupport;
import net.shibboleth.utilities.java.support.logic.Constraint;
import net.shibboleth.utilities.java.support.service.ReloadableService;
import net.shibboleth.utilities.java.support.service.ServiceableComponent;


/** Service that filters out attributes and values based upon loaded policies. */
@ThreadSafe
public class AttributeFilterImpl extends net.shibboleth.idp.attribute.filter.impl.AttributeFilterImpl implements AttributeFilter {

    /** Class logger. */
    @Nonnull private final Logger log = LoggerFactory.getLogger(AttributeFilterImpl.class);

    /** Transcoder registry service object. */
    @Nonnull private ReloadableService<AttributeTranscoderRegistry> transcoderRegistry;

    /**
     * {@inheritDoc}
     */
    public AttributeFilterImpl(@Nonnull @NotEmpty String engineId,
            @Nullable @NullableElements final Collection<AttributeFilterPolicy> policies,
            @Nonnull final ReloadableService<AttributeTranscoderRegistry> registry) {
        super(engineId, policies);
        transcoderRegistry = registry;
    }

    /**
     * {@inheritDoc}
     */
    @Override public void filterAttributes(@Nonnull final AttributeFilterContext filterContext)
            throws AttributeFilterException {
        ComponentSupport.ifNotInitializedThrowUninitializedComponentException(this);
        ComponentSupport.ifDestroyedThrowDestroyedComponentException(this);

        Constraint.isNotNull(filterContext, "Attribute filter context can not be null");
        Map<String, IdPAttribute> prefilteredAttributes = filterContext.getPrefilteredIdPAttributes();
    
        // Create work context to hold intermediate results.
        final AttributeFilterWorkContext filterWorkContextIncludedOption =
            filterContext.getSubcontext(AttributeFilterWorkContext.class, true);

        // Create optional context to hold intermediate results.
        final AttributeFilterOptionalContext filterOptionalContext = 
            filterContext.getSubcontext(AttributeFilterOptionalContext.class, true);
    
        log.debug("{} Beginning process of filtering the following {} attributes: {}", new Object[] {getLogPrefix(),
                prefilteredAttributes.size(), prefilteredAttributes.keySet(),});
    
        // Apply policies with optional values.
        final List<AttributeFilterPolicy> policies = getFilterPolicies();
        for (AttributeFilterPolicy policy : policies) {
            policy.apply(filterContext);
        }

        // Store results at first policy application.
        final Map<String, Set<IdPAttributeValue>> permittedAttributeValueMap =
            filterWorkContextIncludedOption.getPermittedIdPAttributeValues();
        final Map<String, Set<IdPAttributeValue>> deniedAttributeValueMap =
            filterWorkContextIncludedOption.getDeniedAttributeValues();

        // Remove work context at first policy application and create the context again.
        filterContext.removeSubcontext(AttributeFilterWorkContext.class);
        final AttributeFilterWorkContext filterWorkContextExcludedOption =
            filterContext.getSubcontext(AttributeFilterWorkContext.class, true);

        // Set mandatory value mode is on.
        filterOptionalContext.setMandatoryValueMode(true);

        // Apply policies without optional values.
        for (AttributeFilterPolicy policy : policies) {
            policy.apply(filterContext);
        }

        UApproveJPIdPAttribute filteredAttribute;
        for (String attributeId : filterContext.getPrefilteredIdPAttributes().keySet()) {
            log.debug("{} in loop attributeId {}", new Object[] {getLogPrefix(), attributeId,});
            final Collection permittedAttributeValues = permittedAttributeValueMap.get(attributeId);
            if (null == permittedAttributeValues || permittedAttributeValues.isEmpty()) {
                log.debug("Attribute filtering engine '{}': no policy permitted release of attribute {} values", getId(),
                          attributeId);
                continue;
            }

            final Collection mandatoryPermittedAttributeValues = 
                filterWorkContextExcludedOption.getPermittedIdPAttributeValues().get(attributeId);

            // Remove denied values from permitted values.
            final Collection deniedAttributeValues = deniedAttributeValueMap.get(attributeId);
            if (null != deniedAttributeValues && !deniedAttributeValues.isEmpty()) {
                permittedAttributeValues.removeAll(deniedAttributeValues);

                if (null != mandatoryPermittedAttributeValues) {
                    mandatoryPermittedAttributeValues.removeAll(deniedAttributeValues);
                }

                if (permittedAttributeValues.isEmpty()) {
                    log.debug("Attribute filtering engine '{}': deny policies filtered out all values for attribute '{}'",
                              getId(), attributeId);
                    continue;
                }
            }
            log.debug("Attribute filtering engine '{}': {} values for attribute '{}' remained after filtering",
                      new Object[] {getId(), permittedAttributeValues.size(), attributeId,});

            /* 
             * When one or more mandatory values exists, mark this attribute as mandatory attribute.
             * Otherwise mark it as optional attribute.
             */
            boolean optional = false;
            if (null == mandatoryPermittedAttributeValues || mandatoryPermittedAttributeValues.isEmpty()) {
                optional = true;
            }
            log.debug("Attribute filtering engine '{}': attribute {} is {}", getId(), attributeId,
                      optional ? "optional" : "mandatory");

            try {
                filteredAttribute = new UApproveJPIdPAttribute(attributeId);
                filteredAttribute.setIdPAttributeProperties(prefilteredAttributes.get(attributeId).clone());
            } catch (final CloneNotSupportedException e) {
                throw new AttributeFilterException(e);
            }

            final Multimap<String,String> mapped = HashMultimap.create();
            ServiceableComponent<AttributeTranscoderRegistry> component = null;
            try {
                component = transcoderRegistry.getServiceableComponent();
                if (component == null) {
                    log.error("Attribute transcoder service unavailable");
                    return;
                }
                getAttributeNames(component.getComponent(), (IdPAttribute)filteredAttribute, mapped);
            } finally {
                if (component != null) {
                    component.unpinComponent();
                }
            }

            filteredAttribute.setValues(permittedAttributeValues);
            filteredAttribute.setOptional(optional);

            for (Entry<String, String> entry : mapped.entries()) {
                log.info("{}  key '{}' value '{}'", filteredAttribute.getId(), entry.getKey(), entry.getValue());
                if (entry.getKey().equals(SAML2AttributeTranscoder.PROP_NAME)) {
                    filteredAttribute.setSAML2AttributeName(entry.getValue());
                } else if (entry.getKey().equals(SAML2AttributeTranscoder.PROP_NAME_FORMAT)) {
                    filteredAttribute.setSAML2AttributeNameFormat(entry.getValue());
                } else if (entry.getKey().equals(SAML1AttributeTranscoder.PROP_NAME)) {
                    filteredAttribute.setSAML1AttributeName(entry.getValue());
                } else if (entry.getKey().equals(SAML1AttributeTranscoder.PROP_NAMESPACE)) {
                    filteredAttribute.setSAML1AttributeNameFormat(entry.getValue());
                }
            }

            filterContext.getFilteredIdPAttributes().put(filteredAttribute.getId(), filteredAttribute);
        }
    }

    /**
     * Get Attribute Name from IdPAttribute
     *
     * @param registry registry service interface
     * @param input the IdPAttribute to get attribute name
     * @param results names
     */
    private void getAttributeNames(@Nonnull final AttributeTranscoderRegistry registry,
        @Nonnull final IdPAttribute input,
        @Nonnull @NonnullElements final Multimap<String,String> results) {

        Collection<TranscodingRule> transcodingRules =
           registry.getTranscodingRules(input, org.opensaml.saml.saml2.core.Attribute.class);
        if (transcodingRules.isEmpty()) {
            log.info("No transcoding SAML2 rule for Attribute '{}'", input.getId());
            return;
        }

        log.trace("{} find SAML2 transcodingRules size {}", new Object[] {getLogPrefix(), transcodingRules.size(),});
        for (final TranscodingRule rules : transcodingRules) {
            final String propName = SAML2AttributeTranscoder.PROP_NAME;
            final String name = rules.get(propName, String.class);
            log.info("{} SAML2 name '{}'", input.getId(), name);
            if (!Strings.isNullOrEmpty(name)) {
                results.put(propName, name);
            }
            final String propNameFormat = SAML2AttributeTranscoder.PROP_NAME_FORMAT;
            final String nameFormat =
                rules.getOrDefault(propNameFormat, String.class, org.opensaml.saml.saml2.core.Attribute.URI_REFERENCE);
            results.put(propNameFormat, nameFormat);
        }

        transcodingRules = registry.getTranscodingRules(input, org.opensaml.saml.saml1.core.Attribute.class);
        if (transcodingRules.isEmpty()) {
            log.info("No transcoding SAML1 rule for Attribute '{}'", input.getId());
            return;
        }

        log.trace("{} find SAML1 transcodingRules size {}", new Object[] {getLogPrefix(), transcodingRules.size(),});
        for (final TranscodingRule rules : transcodingRules) {
            final String propName = SAML1AttributeTranscoder.PROP_NAME;
            final String name = rules.get(propName, String.class);
            log.info("{} SAML1 name '{}'", input.getId(), name);
            if (!Strings.isNullOrEmpty(name)) {
                results.put(propName, name);
            }
            final String propNameFormat = SAML1AttributeTranscoder.PROP_NAMESPACE;
            final String nameFormat =
                rules.getOrDefault(propNameFormat, String.class, SAMLConstants.SAML1_ATTR_NAMESPACE_URI);
            results.put(propNameFormat, nameFormat);
        }
    }
}
