/*
 * Copyright (c) 2016-2019, National Institute of Informatics
 */
/*
 * Licensed to the University Corporation for Advanced Internet Development,
 * Inc. (UCAID) under one or more contributor license agreements.  See the
 * NOTICE file distributed with this work for additional information regarding
 * copyright ownership. The UCAID licenses this file to You under the Apache
 * License, Version 2.0 (the "License"); you may not use this file except in
 * compliance with the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package jp.gakunin.idp.attribute.filter.matcher.saml.impl;

import java.time.Instant;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;

import org.opensaml.core.xml.XMLObject;
import org.opensaml.core.xml.schema.XSAny;
import org.opensaml.core.xml.schema.XSBase64Binary;
import org.opensaml.core.xml.schema.XSBoolean;
import org.opensaml.core.xml.schema.XSDateTime;
import org.opensaml.core.xml.schema.XSInteger;
import org.opensaml.core.xml.schema.XSString;
import org.opensaml.core.xml.schema.XSURI;
import org.opensaml.saml.common.messaging.context.AttributeConsumingServiceContext;
import org.opensaml.saml.common.messaging.context.SAMLMetadataContext;
import org.opensaml.saml.saml2.core.Attribute;
import org.opensaml.saml.saml2.metadata.AttributeConsumingService;
import org.opensaml.saml.saml2.metadata.RequestedAttribute;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.google.common.collect.Multimap;

import net.shibboleth.idp.attribute.AttributesMapContainer;
import net.shibboleth.idp.attribute.IdPAttribute;
import net.shibboleth.idp.attribute.IdPAttributeValue;
import net.shibboleth.idp.attribute.IdPRequestedAttribute;
import net.shibboleth.idp.attribute.StringAttributeValue;
import net.shibboleth.idp.attribute.filter.Matcher;
import net.shibboleth.idp.attribute.filter.context.AttributeFilterContext;
import net.shibboleth.utilities.java.support.annotation.constraint.NonnullElements;
import net.shibboleth.utilities.java.support.annotation.constraint.NotEmpty;
import net.shibboleth.utilities.java.support.annotation.constraint.Unmodifiable;
import net.shibboleth.utilities.java.support.component.AbstractIdentifiableInitializableComponent;
import net.shibboleth.utilities.java.support.component.ComponentSupport;
import net.shibboleth.utilities.java.support.primitive.StringSupport;
import net.shibboleth.utilities.java.support.xml.DOMTypeSupport;

import jp.gakunin.idp.attribute.filter.context.AttributeFilterOptionalContext;
import net.shibboleth.utilities.java.support.logic.Constraint;

import org.opensaml.saml.saml2.core.Attribute;
import org.opensaml.saml.common.messaging.context.AttributeConsumingServiceContext;
import org.opensaml.saml.common.messaging.context.SAMLMetadataContext;
import org.opensaml.saml.saml2.metadata.AttributeConsumingService;
import org.opensaml.saml.saml2.metadata.RequestedAttribute;


/**
 * Matcher that checks whether an attribute is enumerated in an SP's metadata as a required or optional attribute. Also
 * supports simple value filtering.
 */
public class AttributeInMetadataMatcher extends net.shibboleth.idp.attribute.filter.matcher.saml.impl.AttributeInMetadataMatcher implements Matcher {

    /** Class logger. */
    @Nonnull private final Logger log = LoggerFactory.getLogger(AttributeInMetadataMatcher.class);

    /** Whether the attribute is approved. */
    private boolean onlyIfChecked = false;

    /** Whether optionally requested attributes should be matched. */
    private boolean onlyIfRequired = true;

    /** Whether to return a match if the metadata does not contain an ACS descriptor. */
    private boolean matchIfMetadataSilent;
    
    /** The SAML Attribute Name to look for in the metadata. */
    @Nullable @NotEmpty private String attributeName;

    /** The SAML Attribute NameFormat to look for in the metadata. */
    @Nullable @NotEmpty private String attributeNameFormat;
    
    /** The String used to prefix log message. */
    @Nullable @NotEmpty private String logPrefix;

    /**
     * Gets whether the attribute has been approved.
     * 
     * @return Whether the attribute has been approved
     */
    public boolean getOnlyIfChecked() {
        return onlyIfChecked;
    }

    /**
     * Sets whether the attribute has been approved.
     * 
     * @param flag whether the attribute has been approved.
     */
    public void setOnlyIfChecked(final boolean flag) {
        onlyIfChecked = flag;
    }

// Checkstyle: CyclomaticComplexity|ReturnCount|MethodLength OFF
    /** {@inheritDoc} */
    @Override @Nonnull public Set<IdPAttributeValue> getMatchingValues(@Nonnull final IdPAttribute attribute,
            @Nonnull final AttributeFilterContext filterContext) {
        
        final Set<IdPAttributeValue> values = super.getMatchingValues(attribute, filterContext);
        if (values.isEmpty()) {
            return values;
        }

        if (!getOnlyIfChecked()) {
            log.debug("{} onlyIfChecked is false", getLogPrefix());
            return values;
        }

        final AttributeFilterOptionalContext filterOptionalContext =
                filterContext.getSubcontext(AttributeFilterOptionalContext.class, false);
        Constraint.isNotNull(filterOptionalContext, "Attribute filter work context can not be null");

        if (!filterOptionalContext.isMandatoryValueMode()) {
            return values;
        }

        if (isOptional(attribute, filterContext)) {
            log.debug("{} no optional value mode is enabled", getLogPrefix());
            return Collections.emptySet();
        }

        return values;
    
    }
// Checkstyle: CyclomaticComplexity|ReturnCount|MethodLength ON
    
    /**
     * Whether optional attribute.
     *
     * @return boolean Whether optional attribute
     */
    private boolean isOptional(@Nonnull final IdPAttribute attribute,
            @Nonnull final AttributeFilterContext filterContext) {
        final AttributeConsumingService service = getAttributeConsumingService(filterContext);
        if (service == null) {
            return true;
        }

        if (attributeName != null) {
            // Look for a RequestedAttribute explicitly identified by config.
            RequestedAttribute requestedAttribute = null;
            requestedAttribute = findInMetadata(service, attributeName, attributeNameFormat);
            if (requestedAttribute == null) {
                return false;
            }

            if (requestedAttribute.isRequired()) {
                return false;
            }

            return true;
        }

        // Utilize pre-mapped approach.
        final List<AttributesMapContainer> containerList =
                service.getObjectMetadata().get(AttributesMapContainer.class);
        if (null == containerList || containerList.isEmpty() || containerList.get(0).get() == null ||
                containerList.get(0).get().isEmpty()) {
            log.debug("{} No decoded attributes found when filtering", getLogPrefix());
            log.debug("{} The peer's metadata did not contain requested attribute information"
                    + ", returning false", getLogPrefix());
            return false;
        }
        if (containerList.size() > 1) {
            log.error("{} More than one set of mapped attributes found when filtering, this shouldn't ever happen",
                    getLogPrefix());
        }

        final Multimap<String,IdPAttribute> requestedAttributes = containerList.get(0).get();

        final Collection<? extends IdPAttribute> requestedAttributeList =
                requestedAttributes.get(attribute.getId());

        IdPRequestedAttribute requestedAttribute = null;
        for (final IdPAttribute attr : List.copyOf(requestedAttributeList)) {
            if (attr instanceof IdPRequestedAttribute) {
                requestedAttribute = (IdPRequestedAttribute)attr;
                break;
            }
        }

        if (requestedAttribute == null) {
            return false;
        }

        if (requestedAttribute.isRequired()) {
            return false;
        }

        return true;
    }

    /**
     * This method is just copied from {@net.shibboleth.idp.attribute.filter.matcher.saml.impl.AttributeInMetadataMatcher}
     * because the method is private and we can't call it.
     *
     * Get the appropriate {@link AttributeConsumingService}.
     * 
     * @param filterContext the context for the operation
     * 
     * @return the service, or null
     */
    @Nullable private AttributeConsumingService getAttributeConsumingService(
            @Nonnull final AttributeFilterContext filterContext) {
        
        final SAMLMetadataContext metadataContext = filterContext.getRequesterMetadataContext();
        if (null == metadataContext) {
            log.warn("{} No metadata context when filtering", getLogPrefix());
            return null;
        }
        
        final AttributeConsumingServiceContext acsContext =
                metadataContext.getSubcontext(AttributeConsumingServiceContext.class);
        return acsContext != null ? acsContext.getAttributeConsumingService() : null;
    }
    
    /**
     * This method is just copied from {@net.shibboleth.idp.attribute.filter.matcher.saml.impl.AttributeInMetadataMatcher}
     * because the method is private and we can't call it.
     *
     * Locates a RequestedAttribute object in metadata that matches a specific Attribute Name
     * and NameFormat.
     * 
     * @param service the metadata descriptor to search
     * @param name Attribute Name to match
     * @param nameFormat Attribute NameFormat to match
     * @return a matching RequestedAttribute, or null
     */
    @Nullable private RequestedAttribute findInMetadata(@Nonnull final AttributeConsumingService service,
            @Nonnull final String name, @Nullable final String nameFormat) {
        
        final List<RequestedAttribute> requested = service.getRequestedAttributes();
        for (final RequestedAttribute attr : requested) {
            if (attr.getName().equals(name)) {
                final String format = attr.getNameFormat();
                if (nameFormat == null || format == null || format.equals(Attribute.UNSPECIFIED)
                        || nameFormat.equals(format)) {
                    return attr;
                }
            }
        }
        return null;
    }

}
